@@ -1,6 +1,6 @@
 name = "aidb"
 
-import os, sqlite3, io, gzip, zlib, random, pickle, itertools, warnings, threading
+import os, sqlite3, io, gzip, zlib, random, pickle, itertools, warnings, multiprocessing, h5py
 from itertools import permutations
 from datetime import datetime
 from time import sleep
@@ -17,8 +17,17 @@ import pandas as pd
 import numpy as np
 
 from sklearn.model_selection import train_test_split, StratifiedKFold
+from sklearn.metrics import *
 from sklearn.preprocessing import *
 
+from tqdm import tqdm
+
+import keras 
+from keras.models import Sequential, load_model
+from keras.callbacks import History
+
+import plotly.express as px
+
 # Assumes `pds.create_config()` is run prior to `pds.get_config()`.
 from pydatasci import get_config
 
@@ -39,7 +48,6 @@ def get_db():
 		print("\n Error - Cannot fetch database because it has not yet been configured.\n")
 	else:
 		# peewee ORM connection to database:
-		#db = SqliteDatabase(path)
 		db = SqliteExtDatabase(path)
 		return db
 
@@ -89,7 +97,7 @@ def create_db():
 			Dataset, Label, Featureset, 
 			Splitset, Foldset, Fold, 
 			Algorithm, Hyperparamset, Hyperparamcombo, Preprocess, 
-			Batch, Job
+			Batch, Job, Result
 		])
 		tables = db.get_tables()
 		table_count = len(tables)
@@ -205,7 +213,8 @@ class Dataset(BaseModel):
 			raise ValueError("\nYikes - The dataframe you provided is empty according to `df.empty`")
 
 		Dataset.check_file_format(file_format)
-		Dataset.check_column_count(user_columns=rename_columns, structure=dataframe)
+		if rename_columns is not None:
+			Dataset.check_column_count(user_columns=rename_columns, structure=dataframe)
 
 		shape = {}
 		shape['rows'], shape['columns'] = dataframe.shape[0], dataframe.shape[1]
@@ -261,7 +270,8 @@ class Dataset(BaseModel):
 		, dtype:str = None #pd.Dataframe param
 	):
 		Dataset.check_file_format(file_format)
-		Dataset.check_column_count(user_columns=column_names, structure=ndarray)
+		if rename_columns is not None:
+			Dataset.check_column_count(user_columns=column_names, structure=ndarray)
 
 		if ndarray.size == 0:
 			raise ValueError("\nYikes - The ndarray you provided is empty: `ndarray.size == 0`.\n")
@@ -923,8 +933,6 @@ class Foldset(BaseModel):
 			if (0 > fold_index) or (fold_index > fold_count):
 				raise ValueError("\nYikes - This Foldset <id:" + str(id) +  "> has fold indices between 0 and " + str(fold_count - 1) + "\n")
 
-
-
 		s = foldset.splitset
 		supervision = s.supervision
 		featureset = s.featureset
@@ -1042,10 +1050,13 @@ class Algorithm(BaseModel):
 	# I guess it would be easier to throw 2 models into the mix though.
 	# pytorch and mxnet handle optimizer/loss outside the model definition as part of the train.
 	"""
-	description = CharField(null=True)
+	library = CharField()
+	analysis_type = CharField()#classification_multi, classification_binary, clustering.
 	function_model_build = PickleField()
 	function_model_train = PickleField()
-	function_model_evaluate = PickleField()
+	function_model_predict = PickleField()
+	function_model_loss = PickleField() # do clustering algs have loss?
+	description = CharField(null=True)
 
 
 
@@ -1140,6 +1151,9 @@ class Batch(BaseModel):
 	hyperparamset = ForeignKeyField(Hyperparamset, deferrable='INITIALLY DEFERRED', null=True, backref='batches')
 	foldset = ForeignKeyField(Foldset, deferrable='INITIALLY DEFERRED', null=True, backref='batches')
 
+	def __init__(self, *args, **kwargs):
+		super(Batch, self).__init__(*args, **kwargs)
+
 
 	def from_algorithm(
 		algorithm_id:int
@@ -1201,35 +1215,150 @@ class Batch(BaseModel):
 	def run_jobs(id:int, verbose:bool=False):
 		batch = Batch.get_by_id(id)
 		job_count = batch.job_count
-		jobs = batch.jobs
+		# Want succeeded jobs to appear first so that they get skipped over during a resumed run. Otherwise the % done jumps around.
+		jobs = Job.select().join(Batch).where(Batch.id == batch.id).order_by(Job.status.desc())
 
-		thread_name = "aidb_batch_" + str(batch.id)
-		thread_names = [t.name for t in threading.enumerate()]
-		if thread_name in thread_names:
-			raise ValueError("\nYikes - Cannot start this Batch because it is already running.\n")
+		statuses = Batch.get_statuses(id=batch.id)
+		all_succeeded = all(i == "Succeeded" for i in statuses.values())
+		if all_succeeded:
+			print("\nAll jobs are already complete.\n")
+		elif not (all_succeeded) and ("Succeeded" in statuses.values()):
+			print("\nResuming jobs...\n")
+
+		proc_name = "aidb_batch_" + str(batch.id)
+		proc_names = [p.name for p in multiprocessing.active_children()]
+		if proc_name in proc_names:
+			raise ValueError("\nYikes - Cannot start this Batch because multiprocessing.Process.name '" + proc_name + "' is already running.\n")
 
 		statuses = Batch.get_statuses(id)
 		all_not_started = (set(statuses.values()) == {'Not yet started'})
 		if all_not_started:
 			Job.update(status="Queued").where(Job.batch == id).execute()
 
-		def batch_jobs(): #<-- did not run in background when I use an arg.
-			for j in jobs:
+
+		def background_proc():
+			BaseModel._meta.database.close()
+			BaseModel._meta.database = get_db()
+			for j in tqdm(
+				jobs
+				, desc = "🔮 Training Models 🔮"
+				, ncols = 100
+			):
 				j.run(verbose=verbose)
 
-		t = threading.Thread(target=batch_jobs, name=thread_name)
-		t.start()
-
-		"""
-		if verbose:
-			print("\nTotal jobs to run: " + str(job_count) + "\nstarting queue...")
-		for j in jobs:
-			j.run(verbose=verbose)
-		"""
+		proc = multiprocessing.Process(
+			target = background_proc
+			, name = proc_name
+			, daemon = True
+		)
+		proc.start()
 
 
 	def stop_jobs(id:int):
-		pass
+		# SQLite is ACID (D = Durable) where if a transaction is interrupted it is rolled back.
+		batch = Batch.get_by_id(id)
+		
+		proc_name = "aidb_batch_" + str(batch.id)
+		proc_names = [p.name for p in multiprocessing.active_children()]
+		if proc_name not in proc_names:
+			raise ValueError("\nYikes - Cannot terminate `multiprocessing.Process.name` '" + proc_name + "' because it is not running.\n")
+
+		processes = multiprocessing.active_children()
+		for p in processes:
+			if p.name == proc_name:
+				try:
+					p.terminate()
+				except:
+					raise Exception("\nYikes - Failed to terminate `multiprocessing.Process` '" + proc_name + ".'\n")
+				else:
+					print("\nKilled `multiprocessing.Process` '" + proc_name + "' spawned from Batch <id:" + str(batch.id) + ">.\n")
+
+
+	def metrics_to_pandas(id:int):
+		metric_dicts = Result.select(
+			Result.id, Result.metrics
+		).join(Job).join(Batch).where(Batch.id == id).dicts()
+
+		job_metrics = []
+		# The metrics of each split are grouped under the job id.
+		# Here we break them out so that each split is labeled with its own job id.
+		for d in metric_dicts:
+			for split, data in d['metrics'].items():
+				split_metrics = {}
+				split_metrics['job_id'] = d['id']
+				split_metrics['split'] = split
+
+				for k, v in data.items():
+					split_metrics[k] = v
+
+				job_metrics.append(split_metrics)
+
+		df = pd.DataFrame.from_records(job_metrics)
+		return df
+
+
+	def plot_performance(id:int, max_loss:float=3.0, min_metric_2:float=0.0):
+		batch = Batch.get_by_id(id)
+		analysis_type = batch.algorithm.analysis_type
+		
+		df = batch.metrics_to_pandas()
+		# Now we need to filter the df based on the specified criteria.
+		if (analysis_type == 'classification_multi') or (analysis_type == 'classification_binary'):
+			metric_2 = "accuracy"
+			metric_2_display = "Accuracy"
+		elif analysis_type == 'regression':
+			metric_2 = "r2"
+			metric_2_display = "R²"
+		qry_str = "(loss >= {}) | ({} <= {})".format(max_loss, metric_2, min_metric_2)
+
+		failed = df.query(qry_str)
+		failed_jobs = failed['job_id'].to_list()
+		failed_jobs_unique = list(set(failed_jobs))
+		# Here the `~` inverts it to mean `.isNotIn()`
+		df_passed = df[~df['job_id'].isin(failed_jobs_unique)]
+		df_passed = df_passed.round(3)
+
+		if df_passed.empty:
+			print("There are no models that met the criteria specified.")
+		else:
+			fig = px.line(
+				df_passed
+				, title = '<i>Models Metrics by Split</i>'
+				, x = 'loss'
+				, y = metric_2
+				, color = 'job_id'
+				, height = 600
+				, hover_data = ['job_id', 'split', 'loss', metric_2]
+				, line_shape='spline'
+			)
+			fig.update_traces(
+				mode = 'markers+lines'
+				, line = dict(width = 2)
+				, marker = dict(
+					size = 8
+					, line = dict(
+						width = 2
+						, color = 'white'
+					)
+				)
+			)
+			fig.update_layout(
+				xaxis_title = "Loss"
+				, yaxis_title = metric_2_display
+				, font_family = "Avenir"
+				, font_color = "#FAFAFA"
+				, plot_bgcolor = "#181B1E"
+				, paper_bgcolor = "#181B1E"
+				, hoverlabel = dict(
+					bgcolor = "#0F0F0F"
+					, font_size = 15
+					, font_family = "Avenir"
+				)
+			)
+			fig.update_xaxes(zeroline=False, gridcolor='#262B2F', tickfont=dict(color='#818487'))
+			fig.update_yaxes(zeroline=False, gridcolor='#262B2F', tickfont=dict(color='#818487'))
+			fig.show()
+
 
 
 
@@ -1244,134 +1373,518 @@ class Job(BaseModel):
 	batch = ForeignKeyField(Batch, backref='jobs')
 	hyperparamcombo = ForeignKeyField(Hyperparamcombo, deferrable='INITIALLY DEFERRED', null=True, backref='jobs')
 	fold = ForeignKeyField(Fold, deferrable='INITIALLY DEFERRED', null=True, backref='jobs')
-	#preproc
 
-	# split into sub functions?
-	def run(id:int, verbose:bool=True):
+
+	def split_classification_metrics(labels_processed, predictions, probabilities, analysis_type):
+		if analysis_type == "classification_binary":
+			average = "binary"
+			roc_average = "micro"
+			roc_multi_class = None
+		elif analysis_type == "classification_multi":
+			average = "weighted"
+			roc_average = "weighted"
+			roc_multi_class = "ovr"
+			
+		split_metrics = {}
+		# Let the classification_multi labels hit this metric in OHE format.
+		split_metrics['roc_auc'] = roc_auc_score(labels_processed, probabilities, average=roc_average, multi_class=roc_multi_class)
+		# Then convert the classification_multi labels ordinal format.
+		if analysis_type == "classification_multi":
+			labels_processed = np.argmax(labels_processed, axis=1)
+
+		split_metrics['accuracy'] = accuracy_score(labels_processed, predictions)
+		split_metrics['precision'] = precision_score(labels_processed, predictions, average=average)
+		split_metrics['recall'] = recall_score(labels_processed, predictions, average=average)
+		split_metrics['f1'] = f1_score(labels_processed, predictions, average=average)
+		return split_metrics
+
+
+	def split_regression_metrics(labels, predictions):
+		split_metrics = {}
+		split_metrics['r2'] = r2_score(labels, predictions)
+		split_metrics['mse'] = mean_squared_error(labels, predictions)
+		split_metrics['explained_variance'] = explained_variance_score(labels, predictions)
+		return split_metrics
+
+
+	def split_classification_plots(labels_processed, predictions, probabilities, analysis_type):
+		predictions = predictions.flatten()
+		probabilities = probabilities.flatten()
+		split_plot_data = {}
+		
+		if analysis_type == "classification_binary":
+			labels_processed = labels_processed.flatten()
+			split_plot_data['confusion_matrix'] = confusion_matrix(labels_processed, predictions)
+			fpr, tpr, _ = roc_curve(labels_processed, probabilities)
+			precision, recall, _ = precision_recall_curve(labels_processed, probabilities)
+		
+		elif analysis_type == "classification_multi":
+			# Flatten OHE labels for use with probabilities.
+			labels_flat = labels_processed.flatten()
+			fpr, tpr, _ = roc_curve(labels_flat, probabilities)
+			precision, recall, _ = precision_recall_curve(labels_flat, probabilities)
+
+			# Then convert unflat OHE to ordinal format for use with predictions.
+			labels_ordinal = np.argmax(labels_processed, axis=1)
+			split_plot_data['confusion_matrix'] = confusion_matrix(labels_ordinal, predictions)
+
+		split_plot_data['roc_curve'] = {}
+		split_plot_data['roc_curve']['fpr'] = fpr
+		split_plot_data['roc_curve']['tpr'] = tpr
+		split_plot_data['precision_recall_curve'] = {}
+		split_plot_data['precision_recall_curve']['precision'] = precision
+		split_plot_data['precision_recall_curve']['recall'] = recall
+		return split_plot_data
+
+
+	def run(id:int, verbose:bool=False):
 		j = Job.get_by_id(id)
 		if (j.status == "Succeeded"):
 			if verbose:
-				print("\nSkipping Job #" + str(j.id) + " as is has already been ran.")
+				print("\nSkipping Job #" + str(j.id) + " as is has already succeeded.")
 			return j
 		elif (j.status == "Running"):
 			if verbose:
-				print("\nSkipping Job #" + str(j.id) + " as is has already running.")
+				print("\nSkipping Job #" + str(j.id) + " as it is already running.")
 			return j
 		else:
+			if verbose:
+				print("\nJob #" + str(j.id) + " starting...")
 			algorithm = j.batch.algorithm
+			analysis_type = algorithm.analysis_type
 			splitset = j.batch.splitset
 			preprocess = j.batch.hyperparamset.preprocess
 			hyperparamcombo = j.hyperparamcombo
 			fold = j.fold
 
-			# 1. Fetch training samples.
-			if fold is not None:
-				foldset = fold.foldset
-				fold_samples_np = foldset.to_numpy(fold_index=fold.fold_index)[0]
-
-				samples_train = fold_samples_np['folds_train_combined']
-				# This `samples_evaluate` is just for the training validation history.
-				# We will use the Splitset validation split to formally evaluate later.
-				samples_evaluate = fold_samples_np['fold_validation']
-			else:
-				samples_train = splitset.to_numpy(splits=['train'])['train']
+			# 1. Figure out which splits the model needs to be trained and predicted against. 
+			# The `key_*` variables dynamically determine which splits to use during model_training.
+			# ^ It is being intentionally overwritten as more complex validations/ training splits are introduced.
+			samples = {}
+			if splitset.supervision == "unsupervised":
+				samples['train'] = splitset.to_numpy(splits=['train'])['train']
+				key_train = "train"
+				key_evaluation = None
+			elif splitset.supervision == "supervised":
+				samples['test'] = splitset.to_numpy(splits=['test'])['test']
+				key_evaluation = 'test'
+				
 				if splitset.has_validation:
-					samples_evaluate = splitset.to_numpy(splits=['validation'])['validation']
-				elif splitset.supervision == "supervised":
-					samples_evaluate = splitset.to_numpy(splits=['test'])['test']
-				else:
-					samples_evaluate = None
+					samples['validation'] = splitset.to_numpy(splits=['validation'])['validation']
+					key_evaluation = 'validation'
+					
+				if fold is not None:
+					foldset = fold.foldset
+					fold_samples_np = foldset.to_numpy(fold_index=fold.fold_index)[0]
+					samples['folds_train_combined'] = fold_samples_np['folds_train_combined']
+					samples['fold_validation'] = fold_samples_np['fold_validation']
+					
+					key_train = "folds_train_combined"
+					key_evaluation = "fold_validation"
+				elif fold is None:
+					samples['train'] = splitset.to_numpy(splits=['train'])['train']
+					key_train = "train"
 
 			# 2. Preprocess the features and labels.
+			# Preprocessing happens prior to training the model.
 			if preprocess is not None:
-				# Remember, you only .fit() to training data and then apply transforms to other splits/ folds.
+				# Remember, you only `.fit()` on training data and then apply transforms to other splits/ folds.
 				if preprocess.encoder_features is not None:
 					feature_encoder = preprocess.encoder_features
-					feature_encoder.fit(samples_train['features']) # Hmm. I need this later for test data
-					samples_train['features'] = feature_encoder.transform(samples_train['features'])
+					feature_encoder.fit(samples[key_train]['features'])
+
+					for split, data in samples.items():
+						samples[split]['features'] = feature_encoder.transform(data['features'])
 				
 				if preprocess.encoder_labels is not None:
 					label_encoder = preprocess.encoder_labels
-					label_encoder.fit(samples_train['labels'])
-					samples_train['labels'] = label_encoder.transform(samples_train['labels'])
-				
-				if samples_evaluate is not None:
-					samples_evaluate['features'] = feature_encoder.transform(samples_evaluate['features'])
-					samples_evaluate['labels'] = label_encoder.transform(samples_evaluate['labels'])
+					label_encoder.fit(samples[key_train]['labels'])
+
+					for split, data in samples.items():
+						samples[split]['labels'] = label_encoder.transform(data['labels'])
 
 			# 3. Build and Train model.
 			if hyperparamcombo is not None:
 				hyperparameters = hyperparamcombo.hyperparameters
-			else:
+			elif hyperparamcombo is None:
 				hyperparameters = None
-			
 			model = algorithm.function_model_build(**hyperparameters)
 
 			model = algorithm.function_model_train(
 				model,
-				samples_train,
-				samples_evaluate,
+				samples[key_train],
+				samples[key_evaluation],
 				**hyperparameters
 			)
 
-			# ToDo: create result. write model to Result
-			try:
-				# ToDo: need to let user specify name of history? and weights?
-				# or introduce logic to handle the popular algorithms?
-				model.history
-			except:
-				pass
-			else:
-				# ToDo: update the Result.history
-				pass
+			if (algorithm.library == "Keras"):
+				# If blank this value is `{}` not None.
+				history = model.history.history
 
-			# 4. Fetch remaining evaluation samples. 
-			# We already have `samples_train` and `samples_evaluate` in memory.
-			evaluations = {}
-			evaluations['train'] = algorithm.function_model_evaluate(model, samples_train, **hyperparameters)
+				h5_buffer = io.BytesIO()
+				model.save(
+					h5_buffer
+					, include_optimizer = True
+					, save_format = 'h5'
+				)
+				model_bytes = h5_buffer.getvalue()
 
-			if splitset.supervision == "supervised":
-				# fold_validation.
-				if fold is not None:
-					evaluations['fold_validation'] = algorithm.function_model_evaluate(model, samples_evaluate, **hyperparameters)
-					# at this point for folded sets, validation split may exist. test split exists and needs to be fetched.
-				
-				# validated and unfolded. meaning valiadtion split was already used by `samples_evaluate`.
-				if (splitset.has_validation) and (fold is None):
-					evaluations['validation'] = algorithm.function_model_evaluate(model, samples_evaluate, **hyperparameters)
-				# validated and folded. still need to fetch validation split.
-				elif (splitset.has_validation) and (fold is not None):
-					samples_validation = splitset.to_numpy(splits=['validation'])['validation']
-					if preprocess is not None:
-						samples_validation['features'] = feature_encoder.transform(samples_validation['features'])
-						samples_validation['labels'] = label_encoder.transform(samples_validation['labels'])
-					evaluations['validation'] = algorithm.function_model_evaluate(model, samples_validation, **hyperparameters)
+			# 4. Fetch samples for evaluation.
+			predictions = {}
+			probabilities = {}
+			metrics = {}
+			plot_data = {}
 
-				# unvalidated.
-				if not splitset.has_validation:
-					# then test split was already used by `samples_evaluate`
-					evaluations['test'] = algorithm.function_model_evaluate(model, samples_evaluate, **hyperparameters)
-				else:
-					# otherwise, the test split still needs to be fetched.
-					samples_test = splitset.to_numpy(splits=['test'])['test']
-					if preprocess is not None:
-						samples_test['features'] = feature_encoder.transform(samples_test['features'])
-						samples_test['labels'] = label_encoder.transform(samples_test['labels'])
-					evaluations['test'] = algorithm.function_model_evaluate(model, samples_test, **hyperparameters)
+			if (analysis_type == "classification_multi") or (analysis_type == "classification_binary"):
+				for split, data in samples.items():
+					preds, probs = algorithm.function_model_predict(model, data)
+					predictions[split] = preds
+					probabilities[split] = probs
+
+					metrics[split] = Job.split_classification_metrics(
+						data['labels'], 
+						preds, probs, analysis_type
+					)
+					metrics[split]['loss'] = algorithm.function_model_loss(model, data)
+					plot_data[split] = Job.split_classification_plots(
+						data['labels'], 
+						preds, probs, analysis_type
+					)
+			elif analysis_type == "regression":
+				probabilities = None
+				for split, data in samples.items():
+					preds = algorithm.function_model_predict(model, data)
+					predictions[split] = preds
+					metrics[split] = Job.split_regression_metrics(
+						data['labels'], preds
+					)
+					metrics[split]['loss'] = algorithm.function_model_loss(model, data)
+					plot_data = None
+
+			r = Result.create(
+				model_file = model_bytes
+				, history = history
+				, predictions = predictions
+				, probabilities = probabilities
+				, metrics = metrics
+				, plot_data = plot_data
+				, job = j
+			)
 
-			#ToDo: write evaluations to Result
-			if verbose:
-				print("\nJob #" + str(j.id) + " results: " + str(evaluations))
 			j.status = "Succeeded"
 			j.save()
-			return j #return j.Result
+			return j
+
+
 
 
 class Result(BaseModel):
-	job = ForeignKeyField(Job, backref='results')
-	history = PickleField()
-	model_trained = PickleField() 
+	"""
+	- The classes of encoded labels are all based on train labels.
+	"""
+	model_file = BlobField()
+	history = JSONField()
+	predictions = PickleField()
+	metrics = PickleField()
+	plot_data = PickleField(null=True)
+	probabilities = PickleField(null=True)
 
+	job = ForeignKeyField(Job, backref='results')
+
+
+	def get_model(id:int):
+		r = Result.get_by_id(id)
+		algorithm = r.job.batch.algorithm
+		model_bytes = r.model_file
+		model_bytesio = io.BytesIO(model_bytes)
+		if (algorithm.library == "Keras"):
+			h5_file = h5py.File(model_bytesio,'r')
+			model = load_model(h5_file, compile=True)
+		return model
+
+
+	def plot_learning_curve(id:int):
+		r = Result.get_by_id(id)
+		a = r.job.batch.algorithm
+		analysis_type = a.analysis_type
+
+		history = r.history
+		df = pd.DataFrame.from_dict(history, orient='index').transpose()
+
+		df_loss = df[['loss','val_loss']]
+		df_loss = df_loss.rename(columns={"loss": "train_loss", "val_loss": "validation_loss"})
+		df_loss = df_loss.round(3)
+
+		fig_loss = px.line(
+			df_loss
+			, title = '<i>Training History: Loss</i>'
+			, line_shape = 'spline'
+		)
+		fig_loss.update_layout(
+			xaxis_title = "Epochs"
+			, yaxis_title = "Loss"
+			, legend_title = None
+			, font_family = "Avenir"
+			, font_color = "#FAFAFA"
+			, plot_bgcolor = "#181B1E"
+			, paper_bgcolor = "#181B1E"
+			, height = 400
+			, hoverlabel = dict(
+				bgcolor = "#0F0F0F"
+				, font_size = 15
+				, font_family = "Avenir"
+			)
+			, yaxis = dict(
+				side = "right"
+				, tickmode = 'linear'
+				, tick0 = 0.0
+				, dtick = 0.1
+			)
+			, legend = dict(
+				orientation="h"
+				, yanchor="bottom"
+				, y=1.02
+				, xanchor="right"
+				, x=1
+			)
+			, margin = dict(
+				t = 5
+				, b = 0
+			),
+		)
+		fig_loss.update_xaxes(zeroline=False, gridcolor='#262B2F', tickfont=dict(color='#818487'))
+		fig_loss.update_yaxes(zeroline=False, gridcolor='#262B2F', tickfont=dict(color='#818487'))
+
+		if (analysis_type == "classification_multi") or (analysis_type == "classification_binary"):
+			df_acc = df[['accuracy', 'val_accuracy']]
+			df_acc = df_acc.rename(columns={"accuracy": "train_accuracy", "val_accuracy": "validation_accuracy"})
+			df_acc = df_acc.round(3)
+
+			fig_acc = px.line(
+			df_acc
+				, title = '<i>Training History: Accuracy</i>'
+				, line_shape = 'spline'
+			)
+			fig_acc.update_layout(
+				xaxis_title = "epochs"
+				, yaxis_title = "accuracy"
+				, legend_title = None
+				, font_family = "Avenir"
+				, font_color = "#FAFAFA"
+				, plot_bgcolor = "#181B1E"
+				, paper_bgcolor = "#181B1E"
+				, height = 400
+				, hoverlabel = dict(
+					bgcolor = "#0F0F0F"
+					, font_size = 15
+					, font_family = "Avenir"
+				)
+				, yaxis = dict(
+				side = "right"
+				, tickmode = 'linear'
+				, tick0 = 0.0
+				, dtick = 0.05
+				)
+				, legend = dict(
+					orientation="h"
+					, yanchor="bottom"
+					, y=1.02
+					, xanchor="right"
+					, x=1
+				)
+				, margin = dict(
+					t = 5
+				),
+			)
+			fig_acc.update_xaxes(zeroline=False, gridcolor='#262B2F', tickfont=dict(color='#818487'))
+			fig_acc.update_yaxes(zeroline=False, gridcolor='#262B2F', tickfont=dict(color='#818487'))
+			fig_acc.show()
+		fig_loss.show()
+		
+
+	
+
+	def plot_confusion_matrix(id:int):
+		r = Result.get_by_id(id)
+		result_plot_data = r.plot_data
+		a = r.job.batch.algorithm
+		analysis_type = a.analysis_type
+		if analysis_type == "regression":
+			raise ValueError("\nYikes - `Algorith.analysis_type` of 'regression' does not support this chart.\n")
+		
+
+		cm_by_split = {}
+		for split, data in result_plot_data.items():
+			cm_by_split[split] = data['confusion_matrix']
+		
+		for split, cm in cm_by_split.items():
+			fig = px.imshow(
+				cm
+				, color_continuous_scale = px.colors.sequential.BuGn
+				, labels=dict(x="Predicted Label", y="Actual Label")
+			)
+			fig.update_layout(
+				title = "<i>Confusion Matrix: " + split + "</i>"
+				, xaxis_title = "Predicted Label"
+				, yaxis_title = "Actual Label"
+				, legend_title = 'Sample Count'
+				, font_family = "Avenir"
+				, font_color = "#FAFAFA"
+				, plot_bgcolor = "#181B1E"
+				, paper_bgcolor = "#181B1E"
+				, height = 225 # if too small, it won't render in Jupyter.
+				, hoverlabel = dict(
+					bgcolor = "#0F0F0F"
+					, font_size = 15
+					, font_family = "Avenir"
+				)
+				, yaxis = dict(
+					tickmode = 'linear'
+					, tick0 = 0.0
+					, dtick = 1.0
+				)
+				, margin = dict(
+					b = 0
+					, t = 75
+				)
+			)
+			fig.show()
+
+
+	def plot_precision_recall(id:int):
+		r = Result.get_by_id(id)
+		result_plot_data = r.plot_data
+		a = r.job.batch.algorithm
+		analysis_type = a.analysis_type
+		if analysis_type == "regression":
+			raise ValueError("\nYikes - `Algorith.analysis_type` of 'regression' does not support this chart.\n")
+
+		pr_by_split = {}
+		for split, data in result_plot_data.items():
+			pr_by_split[split] = data['precision_recall_curve']
+
+		dfs = []
+		for split, data in pr_by_split.items():
+			df = pd.DataFrame()
+			df['precision'] = pd.Series(pr_by_split[split]['precision'])
+			df['recall'] = pd.Series(pr_by_split[split]['recall'])
+			df['split'] = split
+			dfs.append(df)
+		dfs = pd.concat(dfs, ignore_index=True)
+		dfs = dfs.round(3)
+
+		fig = px.line(
+			dfs
+			, x = 'recall'
+			, y = 'precision'
+			, color = 'split'
+			, title = '<i>Precision-Recall Curves</i>'
+		)
+		fig.update_layout(
+			legend_title = None
+			, font_family = "Avenir"
+			, font_color = "#FAFAFA"
+			, plot_bgcolor = "#181B1E"
+			, paper_bgcolor = "#181B1E"
+			, height = 500
+			, hoverlabel = dict(
+				bgcolor = "#0F0F0F"
+				, font_size = 15
+				, font_family = "Avenir"
+			)
+			, yaxis = dict(
+				side = "right"
+				, tickmode = 'linear'
+				, tick0 = 0.0
+				, dtick = 0.05
+			)
+			, legend = dict(
+				orientation="h"
+				, yanchor="bottom"
+				, y=1.02
+				, xanchor="right"
+				, x=1
+			)
+		)
+		fig.update_xaxes(zeroline=False, gridcolor='#262B2F', tickfont=dict(color='#818487'))
+		fig.update_yaxes(zeroline=False, gridcolor='#262B2F', tickfont=dict(color='#818487'))
+		fig.show()
+
+
+	def plot_roc_curve(id:int):
+		r = Result.get_by_id(id)
+		result_plot_data = r.plot_data
+		a = r.job.batch.algorithm
+		analysis_type = a.analysis_type
+		if analysis_type == "regression":
+			raise ValueError("\nYikes - `Algorith.analysis_type` of 'regression' does not support this chart.\n")
+
+		roc_by_split = {}
+		for split, data in result_plot_data.items():
+			roc_by_split[split] = data['roc_curve']
+
+		dfs = []
+		for split, data in roc_by_split.items():
+			df = pd.DataFrame()
+			df['fpr'] = pd.Series(roc_by_split[split]['fpr'])
+			df['tpr'] = pd.Series(roc_by_split[split]['tpr'])
+			df['split'] = split
+			dfs.append(df)
+
+		dfs = pd.concat(dfs, ignore_index=True)
+		dfs = dfs.round(3)
+
+		fig = px.line(
+			dfs
+			, x = 'fpr'
+			, y = 'tpr'
+			, color = 'split'
+			, title = '<i>Receiver Operating Characteristic (ROC) Curves</i>'
+			#, line_shape = 'spline'
+		)
+		fig.update_layout(
+			legend_title = None
+			, font_family = "Avenir"
+			, font_color = "#FAFAFA"
+			, plot_bgcolor = "#181B1E"
+			, paper_bgcolor = "#181B1E"
+			, height = 500
+			, hoverlabel = dict(
+				bgcolor = "#0F0F0F"
+				, font_size = 15
+				, font_family = "Avenir"
+			)
+			, xaxis = dict(
+				title = "False Positive Rate (FPR)"
+				, tick0 = 0.00
+				, range = [-0.025,1]
+			)
+			, yaxis = dict(
+				title = "True Positive Rate (TPR)"
+				, side = "left"
+				, tickmode = 'linear'
+				, tick0 = 0.00
+				, dtick = 0.05
+				, range = [0,1.05]
+			)
+			, legend = dict(
+				orientation="h"
+				, yanchor="bottom"
+				, y=1.02
+				, xanchor="right"
+				, x=1
+			)
+			, shapes=[
+				dict(
+					type = 'line'
+					, y0=0, y1=1
+					, x0=0, x1=1
+					, line = dict(dash='dot', width=2, color='#3b4043')
+			)]
+		)
+		fig.update_xaxes(zeroline=False, gridcolor='#262B2F', tickfont=dict(color='#818487'))
+		fig.update_yaxes(zeroline=False, gridcolor='#262B2F', tickfont=dict(color='#818487'))
+		fig.show()
 
 
 """
